# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest.mock import patch

from cinder import context
from cinder import exception
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import powerflex
from cinder.tests.unit.volume.drivers.dell_emc.powerflex import mocks
from cinder.volume import volume_types


class TestManageExistingSnapshot(powerflex.TestPowerFlexDriver):
    """Test cases for ``PowerFlexDriver.manage_existing_snapshot()``"""

    def setUp(self):
        """Setup a test case environment.

        Creates a fake volume object and sets up the required API responses.
        """
        super(TestManageExistingSnapshot, self).setUp()
        ctx = context.RequestContext('fake', 'fake', auth_token=True)
        self.volume = fake_volume.fake_volume_obj(
            ctx, **{'provider_id': fake.PROVIDER_ID})
        self.snapshot = fake_snapshot.fake_snapshot_obj(
            ctx, **{'provider_id': fake.PROVIDER2_ID})
        self.snapshot2 = fake_snapshot.fake_snapshot_obj(
            ctx, **{'provider_id': fake.PROVIDER3_ID})
        self.snapshot.volume = self.snapshot2.volume = self.volume
        self.snapshot['volume_type_id'] = fake.VOLUME_TYPE_ID
        self.snapshot2['volume_type_id'] = fake.VOLUME_TYPE_ID
        self.snapshot_attached = fake_snapshot.fake_snapshot_obj(
            ctx, **{'provider_id': fake.PROVIDER4_ID})

        self.HTTPS_MOCK_RESPONSES = {
            self.RESPONSE_MODE.Valid: {
                'instances/Volume::' + self.volume['provider_id']:
                    mocks.MockHTTPSResponse({
                        'id': fake.PROVIDER_ID,
                        'sizeInKb': 8388608,
                        'mappedSdcInfo': None,
                        'ancestorVolumeId': None
                    }, 200),
                'instances/Volume::' + self.snapshot['provider_id']:
                    mocks.MockHTTPSResponse({
                        'id': fake.PROVIDER2_ID,
                        'sizeInKb': 8000000,
                        'mappedSdcInfo': None,
                        'ancestorVolumeId': fake.PROVIDER_ID
                    }, 200),
                'instances/Volume::' + self.snapshot2['provider_id']:
                    mocks.MockHTTPSResponse({
                        'id': fake.PROVIDER3_ID,
                        'sizeInKb': 8388608,
                        'mappedSdcInfo': None,
                        'ancestorVolumeId': fake.PROVIDER2_ID
                    }, 200)
            },
            self.RESPONSE_MODE.BadStatus: {
                'instances/Volume::' + self.snapshot['provider_id']:
                    mocks.MockHTTPSResponse({
                        'errorCode': 401,
                        'message': 'BadStatus Volume Test',
                    }, 401),
                'instances/Volume::' + self.snapshot2['provider_id']:
                    mocks.MockHTTPSResponse({
                        'id': fake.PROVIDER3_ID,
                        'sizeInKb': 8388608,
                        'ancestorVolumeId': fake.PROVIDER2_ID
                    }, 200),
                'instances/Volume::' + self.snapshot_attached['provider_id']:
                    mocks.MockHTTPSResponse({
                        'id': fake.PROVIDER4_ID,
                        'sizeInKb': 8388608,
                        'mappedSdcInfo': 'Mapped',
                        'ancestorVolumeId': fake.PROVIDER_ID
                    }, 200)
            }
        }

    def test_no_source_id(self):
        existing_ref = {'source-name': 'scaleioSnapName'}
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot, self.snapshot,
                          existing_ref)

    @patch.object(
        volume_types,
        'get_volume_type',
        return_value={'extra_specs': {'volume_backend_name': 'ScaleIO'}})
    def test_snapshot_not_found(self, _mock_volume_type):
        existing_ref = {'source-id': fake.PROVIDER2_ID}
        self.set_https_response_mode(self.RESPONSE_MODE.BadStatus)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.manage_existing_snapshot, self.snapshot,
                          existing_ref)

    @patch.object(
        volume_types,
        'get_volume_type',
        return_value={'extra_specs': {'volume_backend_name': 'ScaleIO'}})
    def test_snapshot_attached(self, _mock_volume_type):
        self.snapshot_attached['volume_type_id'] = fake.VOLUME_TYPE_ID
        existing_ref = {'source-id': fake.PROVIDER4_ID}
        self.set_https_response_mode(self.RESPONSE_MODE.BadStatus)
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          self.snapshot_attached, existing_ref)

    @patch.object(
        volume_types,
        'get_volume_type',
        return_value={'extra_specs': {'volume_backend_name': 'ScaleIO'}})
    def test_different_ancestor(self, _mock_volume_type):
        existing_ref = {'source-id': fake.PROVIDER3_ID}
        self.set_https_response_mode(self.RESPONSE_MODE.Valid)
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_snapshot,
                          self.snapshot2, existing_ref)

    @patch.object(
        volume_types,
        'get_volume_type',
        return_value={'extra_specs': {'volume_backend_name': 'ScaleIO'}})
    def test_manage_snapshot_get_size_calc(self, _mock_volume_type):
        existing_ref = {'source-id': fake.PROVIDER2_ID}
        self.set_https_response_mode(self.RESPONSE_MODE.Valid)
        result = self.driver.manage_existing_snapshot_get_size(
            self.snapshot, existing_ref)
        self.assertEqual(8, result)

    @patch.object(
        volume_types,
        'get_volume_type',
        return_value={'extra_specs': {'volume_backend_name': 'ScaleIO'}})
    def test_manage_existing_snapshot_valid(self, _mock_volume_type):
        existing_ref = {'source-id': fake.PROVIDER2_ID}
        result = self.driver.manage_existing_snapshot(
            self.snapshot, existing_ref)
        self.assertEqual(fake.PROVIDER2_ID, result['provider_id'])



#!/usr/bin/env python
# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Cron script to generate usage notifications for volumes existing during
the audit period.

Together with the notifications generated by volumes
create/delete/resize, over that time period, this allows an external
system consuming usage notification feeds to calculate volume usage
for each tenant.

Time periods are specified as 'hour', 'month', 'day' or 'year'

- `hour` - previous hour. If run at 9:07am, will generate usage for
  8-9am.
- `month` - previous month. If the script is run April 1, it will
  generate usages for March 1 through March 31.
- `day` -  previous day. if run on July 4th, it generates usages for
  July 3rd.
- `year` - previous year. If run on Jan 1, it generates usages for
  Jan 1 through Dec 31 of the previous year.

"""

import datetime
import sys

import iso8601
from oslo_config import cfg
from oslo_log import log as logging

from cinder import i18n  # noqa
i18n.enable_lazy()
from cinder import context
from cinder.i18n import _
from cinder import objects
from cinder import rpc
from cinder import utils
from cinder import version
import cinder.volume.volume_utils


CONF = cfg.CONF
script_opts = [
    cfg.StrOpt('start_time',
               help="If this option is specified then the start time "
                    "specified is used instead of the start time of the "
                    "last completed audit period."),
    cfg.StrOpt('end_time',
               help="If this option is specified then the end time "
                    "specified is used instead of the end time of the "
                    "last completed audit period."),
    cfg.BoolOpt('send_actions',
                default=False,
                help="Send the volume and snapshot create and delete "
                     "notifications generated in the specified period."),
]
CONF.register_cli_opts(script_opts)


def _time_error(LOG, begin, end):
    if CONF.start_time:
        begin = datetime.datetime.strptime(CONF.start_time,
                                           "%Y-%m-%d %H:%M:%S")
    if CONF.end_time:
        end = datetime.datetime.strptime(CONF.end_time,
                                         "%Y-%m-%d %H:%M:%S")
    begin = begin.replace(tzinfo=iso8601.UTC)
    end = end.replace(tzinfo=iso8601.UTC)
    if end <= begin:
        msg = _("The end time (%(end)s) must be after the start "
                "time (%(start)s).") % {'start': begin,
                                        'end': end}
        LOG.error(msg)
        sys.exit(-1)
    return begin, end


def _vol_notify_usage(LOG, volume_ref, extra_info, admin_context):
    """volume_ref notify usage"""
    try:
        LOG.debug("Send exists notification for <volume_id: "
                  "%(volume_id)s> <project_id %(project_id)s> "
                  "<%(extra_info)s>",
                  {'volume_id': volume_ref.id,
                   'project_id': volume_ref.project_id,
                   'extra_info': extra_info})
        cinder.volume.volume_utils.notify_about_volume_usage(
            admin_context, volume_ref, 'exists', extra_usage_info=extra_info)
    except Exception as exc_msg:
        LOG.error("Exists volume notification failed: %s",
                  exc_msg, resource=volume_ref)


def _snap_notify_usage(LOG, snapshot_ref, extra_info, admin_context):
    """snapshot_ref notify usage"""
    try:
        LOG.debug("Send notification for <snapshot_id: %(snapshot_id)s> "
                  "<project_id %(project_id)s> <%(extra_info)s>",
                  {'snapshot_id': snapshot_ref.id,
                   'project_id': snapshot_ref.project_id,
                   'extra_info': extra_info})
        cinder.volume.volume_utils.notify_about_snapshot_usage(
            admin_context, snapshot_ref, 'exists', extra_info)
    except Exception as exc_msg:
        LOG.error("Exists snapshot notification failed: %s",
                  exc_msg, resource=snapshot_ref)


def _backup_notify_usage(LOG, backup_ref, extra_info, admin_context):
    """backup_ref notify usage"""
    try:
        cinder.volume.volume_utils.notify_about_backup_usage(
            admin_context, backup_ref, 'exists', extra_info)
        LOG.debug("Sent notification for <backup_id: %(backup_id)s> "
                  "<project_id %(project_id)s> <%(extra_info)s>",
                  {'backup_id': backup_ref.id,
                   'project_id': backup_ref.project_id,
                   'extra_info': extra_info})
    except Exception as exc_msg:
        LOG.error("Exists backups notification failed: %s", exc_msg)


def _create_action(obj_ref, admin_context, LOG, notify_about_usage,
                   type_id_str, type_name):
    try:
        local_extra_info = {
            'audit_period_beginning': str(obj_ref.created_at),
            'audit_period_ending': str(obj_ref.created_at),
        }
        LOG.debug("Send create notification for <%(type_id_str)s: %(_id)s> "
                  "<project_id %(project_id)s> <%(extra_info)s>",
                  {'type_id_str': type_id_str,
                   '_id': obj_ref.id,
                   'project_id': obj_ref.project_id,
                   'extra_info': local_extra_info})
        notify_about_usage(admin_context, obj_ref,
                           'create.start', extra_usage_info=local_extra_info)
        notify_about_usage(admin_context, obj_ref,
                           'create.end', extra_usage_info=local_extra_info)
    except Exception as exc_msg:
        LOG.error("Create %(type)s notification failed: %(exc_msg)s",
                  {'type': type_name, 'exc_msg': exc_msg}, resource=obj_ref)


def _delete_action(obj_ref, admin_context, LOG, notify_about_usage,
                   type_id_str, type_name):
    try:
        local_extra_info = {
            'audit_period_beginning': str(obj_ref.deleted_at),
            'audit_period_ending': str(obj_ref.deleted_at),
        }
        LOG.debug("Send delete notification for <%(type_id_str)s: %(_id)s> "
                  "<project_id %(project_id)s> <%(extra_info)s>",
                  {'type_id_str': type_id_str,
                   '_id': obj_ref.id,
                   'project_id': obj_ref.project_id,
                   'extra_info': local_extra_info})
        notify_about_usage(admin_context, obj_ref,
                           'delete.start', extra_usage_info=local_extra_info)
        notify_about_usage(admin_context, obj_ref,
                           'delete.end', extra_usage_info=local_extra_info)
    except Exception as exc_msg:
        LOG.error("Delete %(type)s notification failed: %(exc_msg)s",
                  {'type': type_name, 'exc_msg': exc_msg}, resource=obj_ref)


def _obj_ref_action(_notify_usage, LOG, obj_ref, extra_info, admin_context,
                    begin, end, notify_about_usage, type_id_str, type_name):
    _notify_usage(LOG, obj_ref, extra_info, admin_context)
    if CONF.send_actions:
        if begin < obj_ref.created_at < end:
            _create_action(obj_ref, admin_context, LOG,
                           notify_about_usage, type_id_str, type_name)

        if obj_ref.deleted_at and begin < obj_ref.deleted_at < end:
            _delete_action(obj_ref, admin_context, LOG,
                           notify_about_usage, type_id_str, type_name)


def main():
    objects.register_all()
    admin_context = context.get_admin_context()
    CONF(sys.argv[1:], project='cinder',
         version=version.version_string())
    logging.setup(CONF, "cinder")
    LOG = logging.getLogger("cinder")
    rpc.init(CONF)

    begin, end = utils.last_completed_audit_period()
    begin, end = _time_error(LOG, begin, end)

    LOG.info("Starting volume usage audit")
    LOG.info("Creating usages for %(begin_period)s until %(end_period)s",
             {"begin_period": begin, "end_period": end})

    extra_info = {
        'audit_period_beginning': str(begin),
        'audit_period_ending': str(end),
    }

    volumes = objects.VolumeList.get_all_active_by_window(admin_context,
                                                          begin,
                                                          end)

    LOG.info("Found %d volumes", len(volumes))
    for volume_ref in volumes:
        _obj_ref_action(_vol_notify_usage, LOG, volume_ref, extra_info,
                        admin_context, begin, end,
                        cinder.volume.volume_utils.notify_about_volume_usage,
                        "volume_id", "volume")

    snapshots = objects.SnapshotList.get_all_active_by_window(admin_context,
                                                              begin, end)
    LOG.info("Found %d snapshots", len(snapshots))
    for snapshot_ref in snapshots:
        _obj_ref_action(_snap_notify_usage, LOG, snapshot_ref, extra_info,
                        admin_context, begin, end,
                        cinder.volume.volume_utils.notify_about_snapshot_usage,
                        "snapshot_id", "snapshot")

    backups = objects.BackupList.get_all_active_by_window(admin_context,
                                                          begin, end)

    LOG.info("Found %d backups", len(backups))
    for backup_ref in backups:
        _obj_ref_action(_backup_notify_usage, LOG, backup_ref, extra_info,
                        admin_context, begin, end,
                        cinder.volume.volume_utils.notify_about_backup_usage,
                        "backup_id", "backup")
    LOG.info("Volume usage audit completed")



# Copyright 2017 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Tests for the Inspur InStorage volume driver."""

from unittest import mock

import ddt

from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.inspur.instorage import fakes
from cinder.volume import configuration as conf
from cinder.volume.drivers.inspur.instorage import instorage_common


class CLIParserTestCase(test.TestCase):

    def test_empty(self):
        self.assertEqual(0, len(
            instorage_common.CLIParser('')))
        self.assertEqual(0, len(
            instorage_common.CLIParser(('', 'stderr'))))

    def test_header(self):
        raw = r'''id!name
1!node1
2!node2
'''
        resp = instorage_common.CLIParser(raw, with_header=True)
        self.assertEqual(2, len(resp))
        self.assertEqual('1', resp[0]['id'])
        self.assertEqual('2', resp[1]['id'])

    def test_select(self):
        raw = r'''id!123
name!Bill
name!Bill2
age!30
home address!s1
home address!s2

id! 7
name!John
name!John2
age!40
home address!s3
home address!s4
'''
        resp = instorage_common.CLIParser(raw, with_header=False)
        self.assertEqual([('s1', 'Bill', 's1'), ('s2', 'Bill2', 's2'),
                          ('s3', 'John', 's3'), ('s4', 'John2', 's4')],
                         list(resp.select('home address', 'name',
                                          'home address')))

    def test_lsnode_all(self):
        raw = r'''id!name!UPS_serial_number!WWNN!status
1!node1!!500507680200C744!online
2!node2!!500507680200C745!online
'''
        resp = instorage_common.CLIParser(raw)
        self.assertEqual(2, len(resp))
        self.assertEqual('1', resp[0]['id'])
        self.assertEqual('500507680200C744', resp[0]['WWNN'])
        self.assertEqual('2', resp[1]['id'])
        self.assertEqual('500507680200C745', resp[1]['WWNN'])

    def test_lsnode_single(self):
        raw = r'''id!1
port_id!500507680210C744
port_status!active
port_speed!8Gb
port_id!500507680240C744
port_status!inactive
port_speed!8Gb
'''
        resp = instorage_common.CLIParser(raw, with_header=False)
        self.assertEqual(1, len(resp))
        self.assertEqual('1', resp[0]['id'])
        self.assertEqual([('500507680210C744', 'active'),
                          ('500507680240C744', 'inactive')],
                         list(resp.select('port_id', 'port_status')))


class InStorageAssistantTestCase(test.TestCase):

    def setUp(self):
        super(InStorageAssistantTestCase, self).setUp()
        self.instorage_mcs_common = instorage_common.InStorageAssistant(None)
        self.mock_wait_time = mock.patch.object(
            instorage_common.InStorageAssistant, "WAIT_TIME", 0)

    @mock.patch.object(instorage_common.InStorageSSH, 'lslicense')
    @mock.patch.object(instorage_common.InStorageSSH, 'lsguicapabilities')
    def test_compression_enabled(self, lsguicapabilities, lslicense):
        fake_license_without_keys = {}
        fake_license = {
            'license_compression_enclosures': '1',
            'license_compression_capacity': '1'
        }
        fake_license_scheme = {
            'compression': 'yes'
        }
        fake_license_invalid_scheme = {
            'compression': 'no'
        }

        lslicense.side_effect = [fake_license_without_keys,
                                 fake_license_without_keys,
                                 fake_license,
                                 fake_license_without_keys]
        lsguicapabilities.side_effect = [fake_license_without_keys,
                                         fake_license_invalid_scheme,
                                         fake_license_scheme]
        self.assertFalse(self.instorage_mcs_common.compression_enabled())

        self.assertFalse(self.instorage_mcs_common.compression_enabled())

        self.assertTrue(self.instorage_mcs_common.compression_enabled())

        self.assertTrue(self.instorage_mcs_common.compression_enabled())

    @mock.patch.object(instorage_common.InStorageAssistant,
                       'get_vdisk_count_by_io_group')
    def test_select_io_group(self, get_vdisk_count_by_io_group):
        # given io groups
        opts = {}
        # system io groups
        state = {}

        fake_iog_vdc1 = {0: 100, 1: 50, 2: 50, 3: 300}
        fake_iog_vdc2 = {0: 2, 1: 1, 2: 200}
        fake_iog_vdc3 = {0: 2, 2: 200}
        fake_iog_vdc4 = {0: 100, 1: 100, 2: 100, 3: 100}
        fake_iog_vdc5 = {0: 10, 1: 1, 2: 200, 3: 300}

        get_vdisk_count_by_io_group.side_effect = [fake_iog_vdc1,
                                                   fake_iog_vdc2,
                                                   fake_iog_vdc3,
                                                   fake_iog_vdc4,
                                                   fake_iog_vdc5]
        opts['iogrp'] = '0,2'
        state['available_iogrps'] = [0, 1, 2, 3]

        iog = self.instorage_mcs_common.select_io_group(state, opts)
        self.assertTrue(iog in state['available_iogrps'])
        self.assertEqual(2, iog)

        opts['iogrp'] = '0'
        state['available_iogrps'] = [0, 1, 2]

        iog = self.instorage_mcs_common.select_io_group(state, opts)
        self.assertTrue(iog in state['available_iogrps'])
        self.assertEqual(0, iog)

        opts['iogrp'] = '1,2'
        state['available_iogrps'] = [0, 2]

        iog = self.instorage_mcs_common.select_io_group(state, opts)
        self.assertTrue(iog in state['available_iogrps'])
        self.assertEqual(2, iog)

        opts['iogrp'] = ' 0, 1, 2 '
        state['available_iogrps'] = [0, 1, 2, 3]

        iog = self.instorage_mcs_common.select_io_group(state, opts)
        self.assertTrue(iog in state['available_iogrps'])
        # since vdisk count in all iogroups is same, it will pick the first
        self.assertEqual(0, iog)

        opts['iogrp'] = '0,1,2, 3'
        state['available_iogrps'] = [0, 1, 2, 3]

        iog = self.instorage_mcs_common.select_io_group(state, opts)
        self.assertTrue(iog in state['available_iogrps'])
        self.assertEqual(1, iog)


@ddt.ddt
class InStorageSSHTestCase(test.TestCase):

    def setUp(self):
        super(InStorageSSHTestCase, self).setUp()
        self.fake_driver = fakes.FakeInStorageMCSISCSIDriver(
            configuration=conf.Configuration(None))
        sim = fakes.FakeInStorage(['openstack'])
        self.fake_driver.set_fake_storage(sim)
        self.instorage_ssh = instorage_common.InStorageSSH(
            self.fake_driver._run_ssh)

    def test_mkvdiskhostmap(self):
        # mkvdiskhostmap should not be returning anything
        self.fake_driver.fake_storage._volumes_list['9999'] = {
            'name': ' 9999', 'id': '0', 'uid': '0',
            'IO_group_id': '0', 'IO_group_name': 'fakepool'}
        self.fake_driver.fake_storage._hosts_list['HOST1'] = {
            'name': 'HOST1', 'id': '0', 'host_name': 'HOST1'}
        self.fake_driver.fake_storage._hosts_list['HOST2'] = {
            'name': 'HOST2', 'id': '1', 'host_name': 'HOST2'}
        self.fake_driver.fake_storage._hosts_list['HOST3'] = {
            'name': 'HOST3', 'id': '2', 'host_name': 'HOST3'}

        ret = self.instorage_ssh.mkvdiskhostmap('HOST1', '9999', '511', False)
        self.assertEqual('511', ret)

        ret = self.instorage_ssh.mkvdiskhostmap('HOST2', '9999', '512', True)
        self.assertEqual('512', ret)

        ret = self.instorage_ssh.mkvdiskhostmap('HOST3', '9999', None, True)
        self.assertIsNotNone(ret)

        with mock.patch.object(
                instorage_common.InStorageSSH,
                'run_ssh_check_created') as run_ssh_check_created:
            ex = exception.VolumeBackendAPIException(data='CMMVC6071E')
            run_ssh_check_created.side_effect = ex
            self.assertRaises(exception.VolumeBackendAPIException,
                              self.instorage_ssh.mkvdiskhostmap,
                              'HOST3', '9999', 511, True)

    @ddt.data((exception.VolumeBackendAPIException(data='CMMVC6372W'), None),
              (exception.VolumeBackendAPIException(data='CMMVC6372W'),
               {'name': 'fakevol', 'id': '0', 'uid': '0', 'IO_group_id': '0',
                'IO_group_name': 'fakepool'}),
              (exception.VolumeBackendAPIException(data='error'), None))
    @ddt.unpack
    def test_mkvdisk_with_warning(self, run_ssh_check, lsvol):
        opt = {'iogrp': 0}
        with mock.patch.object(instorage_common.InStorageSSH,
                               'run_ssh_check_created',
                               side_effect=run_ssh_check):
            with mock.patch.object(instorage_common.InStorageSSH, 'lsvdisk',
                                   return_value=lsvol):
                if lsvol:
                    ret = self.instorage_ssh.mkvdisk('fakevol', '1', 'gb',
                                                     'fakepool', opt, [])
                    self.assertEqual('0', ret)
                else:
                    self.assertRaises(exception.VolumeBackendAPIException,
                                      self.instorage_ssh.mkvdisk,
                                      'fakevol', '1', 'gb', 'fakepool',
                                      opt, [])



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from six.moves import http_client

from cinder.api import microversions as mv
from cinder import db
from cinder.tests.unit import fake_constants
from cinder.tests.unit.policies import test_base


class DefaultVolumeTypesPolicyTests(test_base.CinderPolicyTests):

    class FakeDefaultType:
        project_id = fake_constants.PROJECT_ID
        volume_type_id = fake_constants.VOLUME_TYPE_ID

    def setUp(self):
        super(DefaultVolumeTypesPolicyTests, self).setUp()
        self.volume_type = self._create_fake_type(self.admin_context)
        self.project = self.FakeProject()
        # Need to mock out Keystone so the functional tests don't require other
        # services
        _keystone_client = mock.MagicMock()
        _keystone_client.version = 'v3'
        _keystone_client.projects.get.side_effect = self._get_project
        _keystone_client_get = mock.patch(
            'cinder.quota_utils._keystone_client',
            lambda *args, **kwargs: _keystone_client)
        _keystone_client_get.start()
        self.addCleanup(_keystone_client_get.stop)

    def _get_project(self, project_id, *args, **kwargs):
        return self.project

    class FakeProject(object):
        _dom_id = fake_constants.DOMAIN_ID

        def __init__(self, parent_id=None):
            self.id = fake_constants.PROJECT_ID
            self.parent_id = parent_id
            self.domain_id = self._dom_id
            self.subtree = None
            self.parents = None

    def test_system_admin_can_set_default(self):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types/%s' % system_admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        response = self._get_request_response(system_admin_context,
                                              path, 'PUT', body=body,
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.OK, response.status_int)

    def test_project_admin_can_set_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        response = self._get_request_response(admin_context,
                                              path, 'PUT', body=body,
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.OK, response.status_int)

    def test_project_admin_cannot_set_default_for_other_project(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        response = self._get_request_response(self.other_admin_context,
                                              path, 'PUT', body=body,
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    @mock.patch.object(db, 'project_default_volume_type_get',
                       return_value=FakeDefaultType())
    def test_system_admin_can_get_default(self, mock_default_get):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types/%s' % system_admin_context.project_id
        response = self._get_request_response(system_admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.OK, response.status_int)

    def test_project_admin_can_get_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        self._get_request_response(admin_context,
                                   path, 'PUT', body=body,
                                   microversion=
                                   mv.DEFAULT_TYPE_OVERRIDES)

        path = '/v3/default-types/%s' % admin_context.project_id
        response = self._get_request_response(admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.OK, response.status_int)

    def test_project_admin_cannot_get_default_for_other_project(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        response = self._get_request_response(self.other_admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_system_admin_can_get_all_default(self):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types'
        response = self._get_request_response(system_admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.OK, response.status_int)

    def test_project_admin_cannot_get_all_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types'
        response = self._get_request_response(admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.FORBIDDEN, response.status_int)

    def test_system_admin_can_unset_default(self):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types/%s' % system_admin_context.project_id
        response = self._get_request_response(system_admin_context,
                                              path, 'DELETE',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_project_admin_can_unset_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        response = self._get_request_response(admin_context,
                                              path, 'DELETE',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_project_admin_cannot_unset_default_for_other_project(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        response = self._get_request_response(self.other_admin_context,
                                              path, 'DELETE',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(http_client.FORBIDDEN, response.status_int)



#  Copyright (c) 2014-2019 LINBIT HA Solutions GmbH
#  All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""This driver connects Cinder to an installed LINSTOR instance.

See https://docs.linbit.com/docs/users-guide-9.0/#ch-openstack-linstor
for more details.
"""

import socket
import uuid

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import units

from cinder import exception
from cinder.i18n import _
from cinder.image import image_utils
from cinder import interface
from cinder.volume import configuration
from cinder.volume import driver
from cinder.volume import volume_utils

try:
    import linstor
    lin_drv = linstor.Linstor
except ImportError:
    linstor = None
    lin_drv = None

# To override these values, update cinder.conf in /etc/cinder/
linstor_opts = [
    cfg.StrOpt('linstor_default_volume_group_name',
               default='drbd-vg',
               help='Default Volume Group name for LINSTOR. '
                    'Not Cinder Volume.'),

    cfg.StrOpt('linstor_default_uri',
               default='linstor://localhost',
               help='Default storage URI for LINSTOR.'),

    cfg.StrOpt('linstor_default_storage_pool_name',
               default='DfltStorPool',
               help='Default Storage Pool name for LINSTOR.'),

    cfg.FloatOpt('linstor_volume_downsize_factor',
                 default=4096,
                 help='Default volume downscale size in KiB = 4 MiB.'),

    cfg.IntOpt('linstor_default_blocksize',
               default=4096,
               help='Default Block size for Image restoration. '
                    'When using iSCSI transport, this option '
                    'specifies the block size.'),

    cfg.IntOpt('linstor_autoplace_count',
               default=0,
               help='Autoplace replication count on volume deployment. '
                    '0 = Full cluster replication without autoplace, '
                    '1 = Single node deployment without replication, '
                    '2 or greater = Replicated deployment with autoplace.'),

    cfg.BoolOpt('linstor_controller_diskless',
                default=True,
                help='True means Cinder node is a diskless LINSTOR node.')
]

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.register_opts(linstor_opts, group=configuration.SHARED_CONF_GROUP)

CINDER_UNKNOWN = 'unknown'
DM_VN_PREFIX = 'CV_'
DM_SN_PREFIX = 'SN_'
DISKLESS = 'DISKLESS'
LVM = 'LVM'
LVM_THIN = 'LVM_THIN'
ZFS = 'ZFS'
ZFS_THIN = 'ZFS_THIN'


class LinstorBaseDriver(driver.VolumeDriver):
    """Cinder driver that uses LINSTOR for storage.

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.0.1 - Added support for LINSTOR 0.9.12
    """

    VERSION = '1.0.1'

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = 'LINBIT_LINSTOR_CI'

    def __init__(self, *args, **kwargs):
        super(LinstorBaseDriver, self).__init__(*args, **kwargs)
        LOG.debug('START: Base Init Linstor')

        self.configuration.append_config_values(linstor_opts)
        self.default_pool = self.configuration.safe_get(
            'linstor_default_storage_pool_name')
        self.default_uri = self.configuration.safe_get(
            'linstor_default_uri')
        self.default_downsize_factor = self.configuration.safe_get(
            'linstor_volume_downsize_factor')
        self.default_vg_name = self.configuration.safe_get(
            'linstor_default_volume_group_name')
        self.default_blocksize = self.configuration.safe_get(
            'linstor_default_blocksize')
        self.diskless = self.configuration.safe_get(
            'linstor_controller_diskless')
        self.ap_count = self.configuration.safe_get(
            'linstor_autoplace_count')
        self.default_backend_name = self.configuration.safe_get(
            'volume_backend_name')
        self.host_name = socket.gethostname()

    @staticmethod
    def get_driver_options():
        return linstor_opts

    def _ping(self):
        with lin_drv(self.default_uri) as lin:
            return lin.ping()

    def _clean_uuid(self):
        """Returns a UUID string, WITHOUT braces."""
        # Some uuid library versions put braces around the result.
        # We don't want them, just a plain [0-9a-f-]+ string.
        uuid_str = str(uuid.uuid4())
        uuid_str = uuid_str.replace("{", "")
        uuid_str = uuid_str.replace("}", "")
        return uuid_str

    # LINSTOR works in kiB units; Cinder uses GiB.
    def _vol_size_to_linstor(self, size):
        return int(size * units.Mi - self.default_downsize_factor)

    def _vol_size_to_cinder(self, size):
        return int(size / units.Mi)

    def _is_clean_volume_name(self, name, prefix):
        try:
            if (name.startswith(CONF.volume_name_template % "") and
                    uuid.UUID(name[7:]) is not None):
                return prefix + name[7:]
        except ValueError:
            return None

        try:
            if uuid.UUID(name) is not None:
                return prefix + name
        except ValueError:
            return None

    def _snapshot_name_from_cinder_snapshot(self, snapshot):
        sn_name = self._is_clean_volume_name(snapshot['id'], DM_SN_PREFIX)
        return sn_name

    def _cinder_volume_name_from_drbd_resource(self, rsc_name):
        cinder_volume_name = rsc_name.split(DM_VN_PREFIX)[1]
        return cinder_volume_name

    def _drbd_resource_name_from_cinder_snapshot(self, snapshot):
        drbd_resource_name = '{}{}'.format(DM_VN_PREFIX,
                                           snapshot['volume_id'])
        return drbd_resource_name

    def _drbd_resource_name_from_cinder_volume(self, volume):
        drbd_resource_name = '{}{}'.format(DM_VN_PREFIX, volume['id'])
        return drbd_resource_name

    def _get_api_resource_list(self):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()
            api_reply = lin.resource_list()[0].__dict__['_rest_data']
            return api_reply

    def _get_api_resource_dfn_list(self):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()
            api_reply = lin.resource_dfn_list()[0].__dict__['_rest_data']
            return api_reply

    def _get_api_node_list(self):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()
            api_reply = lin.node_list()[0].__dict__['_rest_data']
            return api_reply

    def _get_api_storage_pool_dfn_list(self):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()
            api_reply = lin.storage_pool_dfn_list()[0].__dict__['_rest_data']
            return api_reply

    def _get_api_storage_pool_list(self):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()
            api_reply = lin.storage_pool_list()[0].__dict__['_rest_data']
            return api_reply

    def _get_api_volume_extend(self, rsc_target_name, new_size):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            vol_reply = lin.volume_dfn_modify(
                rsc_name=rsc_target_name,
                volume_nr=0,
                size=self._vol_size_to_linstor(new_size))
            return vol_reply

    def _api_snapshot_create(self, drbd_rsc_name, snapshot_name):
        lin = linstor.Resource(drbd_rsc_name, uri=self.default_uri)
        snap_reply = lin.snapshot_create(snapshot_name)
        return snap_reply

    def _api_snapshot_delete(self, drbd_rsc_name, snapshot_name):
        lin = linstor.Resource(drbd_rsc_name, uri=self.default_uri)
        snap_reply = lin.snapshot_delete(snapshot_name)
        return snap_reply

    def _api_rsc_dfn_delete(self, drbd_rsc_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            snap_reply = lin.resource_dfn_delete(drbd_rsc_name)
            return snap_reply

    def _api_storage_pool_create(self,
                                 node_name,
                                 storage_pool_name,
                                 storage_driver,
                                 driver_pool_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            sp_reply = lin.storage_pool_create(
                node_name=node_name,
                storage_pool_name=storage_pool_name,
                storage_driver=storage_driver,
                driver_pool_name=driver_pool_name)
            return sp_reply

    def _api_rsc_dfn_create(self, rsc_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            rsc_dfn_reply = lin.resource_dfn_create(rsc_name)
            return rsc_dfn_reply

    def _api_volume_dfn_create(self, rsc_name, size):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            vol_dfn_reply = lin.volume_dfn_create(
                rsc_name=rsc_name,
                storage_pool=self.default_pool,
                size=size)
            return vol_dfn_reply

    def _api_volume_dfn_set_sp(self, rsc_target_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            snap_reply = lin.volume_dfn_modify(
                rsc_name=rsc_target_name,
                volume_nr=0,
                set_properties={
                    'StorPoolName': self.default_pool
                })
            return snap_reply

    def _api_rsc_create(self, rsc_name, node_name, diskless=False):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            if diskless:
                storage_pool = None
            else:
                storage_pool = self.default_pool

            new_rsc = linstor.ResourceData(rsc_name=rsc_name,
                                           node_name=node_name,
                                           storage_pool=storage_pool,
                                           diskless=diskless)

            rsc_reply = lin.resource_create([new_rsc], async_msg=False)
            return rsc_reply

    def _api_rsc_autoplace(self, rsc_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            new_rsc = linstor.Resource(name=rsc_name, uri=self.default_uri)
            new_rsc.placement.redundancy = self.ap_count
            new_rsc.placement.storage_pool = self.default_pool
            rsc_reply = new_rsc.autoplace()

            return rsc_reply

    def _api_rsc_delete(self, rsc_name, node_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            rsc_reply = lin.resource_delete(node_name=node_name,
                                            rsc_name=rsc_name)
            return rsc_reply

    def _api_rsc_auto_delete(self, rsc_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            rsc = linstor.Resource(str(rsc_name), self.default_uri)
            return rsc.delete()

    def _api_rsc_is_diskless(self, rsc_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            rsc = linstor.Resource(str(rsc_name))
            return rsc.is_diskless(self.host_name)

    def _api_rsc_size(self, rsc_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            rsc = linstor.Resource(str(rsc_name))
            if len(rsc.volumes):
                if "size" in rsc.volumes:
                    return rsc.volumes[0].size
                else:
                    return 0
            else:
                return 0

    def _api_volume_dfn_delete(self, rsc_name, volume_nr):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            rsc_reply = lin.volume_dfn_delete(rsc_name=rsc_name,
                                              volume_nr=volume_nr)
            return rsc_reply

    def _api_snapshot_volume_dfn_restore(self,
                                         src_rsc_name,
                                         src_snap_name,
                                         new_vol_name):
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            vol_reply = lin.snapshot_volume_definition_restore(
                from_resource=src_rsc_name,
                from_snapshot=src_snap_name,
                to_resource=new_vol_name)
            return vol_reply

    def _api_snapshot_resource_restore(self,
                                       src_rsc_name,
                                       src_snap_name,
                                       new_vol_name):

        lin = linstor.Resource(src_rsc_name, uri=self.default_uri)
        new_rsc = lin.restore_from_snapshot(src_snap_name, new_vol_name)

        # Adds an aux/property KV for synchronous return from snapshot restore
        with lin_drv(self.default_uri) as lin:
            if not lin.connected:
                lin.connect()

            aux_prop = {}
            aux_prop["Aux/restore"] = "done"
            lin.volume_dfn_modify(
                rsc_name=new_vol_name,
                volume_nr=0,
                set_properties=aux_prop)

        if new_rsc.name == new_vol_name:
            return True
        return False

    def _get_rsc_path(self, rsc_name):
        rsc_list_reply = self._get_api_resource_list()

        if rsc_list_reply:
            for rsc in rsc_list_reply:
                if (rsc["name"] == rsc_name and
                        rsc["node_name"] == self.host_name):
                    for volume in rsc["volumes"]:
                        if volume["volume_number"] == 0:
                            return volume["device_path"]

    def _get_local_path(self, volume):
        try:
            full_rsc_name = (
                self._drbd_resource_name_from_cinder_volume(volume))

            return self._get_rsc_path(full_rsc_name)

        except Exception:
            message = _('Local Volume not found.')
            raise exception.VolumeBackendAPIException(data=message)

    def _get_spd(self):
        # Storage Pool Definition List
        spd_list_reply = self._get_api_storage_pool_dfn_list()
        spd_list = []

        if spd_list_reply:
            for spd in spd_list_reply:
                spd_list.append(spd["storage_pool_name"])

        return spd_list

    def _get_storage_pool(self):
        # Fetch Storage Pool List
        sp_list_reply = self._get_api_storage_pool_list()

        # Separate the diskless nodes
        sp_diskless_list = []
        sp_list = []
        node_count = 0

        if sp_list_reply:
            for node in sp_list_reply:
                if node["storage_pool_name"] == self.default_pool:
                    sp_node = {}
                    sp_node["node_name"] = node["node_name"]
                    sp_node["sp_uuid"] = node["uuid"]
                    sp_node["sp_name"] = node["storage_pool_name"]

                    if node["provider_kind"] == DISKLESS:
                        diskless = True
                        sp_node["sp_free"] = -1.0
                        sp_node["sp_cap"] = -1.0
                        sp_node["sp_allocated"] = 0.0
                    else:
                        diskless = False
                        if "free_capacity" in node:
                            temp = float(node["free_capacity"]) / units.Mi
                            sp_node["sp_free"] = round(temp)
                            temp = float(node["total_capacity"]) / units.Mi
                            sp_node["sp_cap"] = round(temp)

                    drivers = [LVM, LVM_THIN, ZFS, ZFS_THIN, DISKLESS]

                    # Driver selection
                    if node["provider_kind"] in drivers:
                        sp_node['driver_name'] = node["provider_kind"]
                    else:
                        sp_node['driver_name'] = str(node["provider_kind"])

                    if diskless:
                        sp_diskless_list.append(sp_node)
                    else:
                        sp_list.append(sp_node)
                    node_count += 1

            # Add the diskless nodes to the end of the list
            if sp_diskless_list:
                sp_list.extend(sp_diskless_list)

        return sp_list

    def _get_volume_stats(self):

        data = {}
        data["volume_backend_name"] = self.default_backend_name
        data["vendor_name"] = "LINBIT"
        data["driver_version"] = self.VERSION
        data["pools"] = []

        sp_data = self._get_storage_pool()
        rd_list = self._get_resource_definitions()

        # Total volumes and capacity
        num_vols = 0
        for rd in rd_list:
            num_vols += 1

        # allocated_sizes_gb = []
        free_gb = []
        total_gb = []
        thin_enabled = False

        # Total & Free capacity for Local Node
        single_pool = {}
        for sp in sp_data:
            if "Diskless" not in sp["driver_name"]:
                thin_backends = [LVM_THIN, ZFS_THIN]
                if sp["driver_name"] in thin_backends:
                    thin_enabled = True
                if "sp_cap" in sp:
                    if sp["sp_cap"] >= 0.0:
                        total_gb.append(sp["sp_cap"])
                if "sp_free" in sp:
                    if sp["sp_free"] >= 0.0:
                        free_gb.append(sp["sp_free"])

        # Allocated capacity
        sp_allocated_size_gb = 0.0
        local_resources = []

        reply = self._get_api_resource_list()

        if reply:
            for rsc in reply:
                if rsc["node_name"] == self.host_name:
                    local_resources.append(rsc["name"])

            for rsc_name in local_resources:
                if not self._api_rsc_is_diskless(rsc_name):
                    rsc_size = self._api_rsc_size(rsc_name)
                    sp_allocated_size_gb += round(
                        int(rsc_size) / units.Gi, 2)

        single_pool["pool_name"] = data["volume_backend_name"]
        single_pool["free_capacity_gb"] = min(free_gb) if free_gb else 0
        single_pool["total_capacity_gb"] = min(total_gb) if total_gb else 0
        single_pool["provisioned_capacity_gb"] = sp_allocated_size_gb
        single_pool["reserved_percentage"] = (
            self.configuration.reserved_percentage)
        single_pool["thin_provisioning_support"] = thin_enabled
        single_pool["thick_provisioning_support"] = not thin_enabled
        single_pool["max_over_subscription_ratio"] = (
            self.configuration.max_over_subscription_ratio)
        single_pool["location_info"] = self.default_uri
        single_pool["total_volumes"] = num_vols
        single_pool["filter_function"] = self.get_filter_function()
        single_pool["goodness_function"] = self.get_goodness_function()
        single_pool["QoS_support"] = False
        single_pool["multiattach"] = False
        single_pool["backend_state"] = "up"

        data["pools"].append(single_pool)

        return data

    def _get_resource_definitions(self):

        rd_list_reply = self._get_api_resource_dfn_list()
        rd_list = []

        if rd_list_reply:
            for node in rd_list_reply:
                # Count only Cinder volumes
                if DM_VN_PREFIX in node['name']:
                    rd_node = {}
                    rd_node["rd_uuid"] = node['uuid']
                    rd_node["rd_name"] = node['name']
                    rd_list.append(rd_node)

        return rd_list

    def _get_snapshot_nodes(self, resource):
        """Returns all available resource nodes for snapshot.

        However, it excludes diskless nodes.
        """
        rsc_list_reply = self._get_api_resource_list()
        snap_list = []

        if rsc_list_reply:
            for rsc in rsc_list_reply:
                if rsc["name"] != resource:
                    continue

                # Diskless nodes are not available for snapshots
                diskless = False
                if "flags" in rsc:
                    if 'DISKLESS' in rsc["flags"]:
                        diskless = True
                if not diskless:
                    snap_list.append(rsc["node_name"])

        return snap_list

    def _get_diskless_nodes(self, resource):
        # Returns diskless nodes given a resource
        rsc_list_reply = self._get_api_resource_list()
        diskless_list = []

        if rsc_list_reply:
            for rsc in rsc_list_reply:
                if rsc["name"] != resource:
                    continue

                if "flags" in rsc:
                    if DISKLESS in rsc["flags"]:
                        diskless_list.append(rsc["node_name"])

        return diskless_list

    def _get_linstor_nodes(self):
        # Returns all available LINSTOR nodes
        node_list_reply = self._get_api_node_list()
        node_list = []

        if node_list_reply:
            for node in node_list_reply:
                node_list.append(node["name"])

        return node_list

    def _get_nodes(self):
        # Returns all LINSTOR nodes in a dict list
        node_list_reply = self._get_api_node_list()
        node_list = []

        if node_list_reply:
            for node in node_list_reply:
                node_item = {}
                node_item["node_name"] = node["name"]
                node_item["node_address"] = (
                    node["net_interfaces"][0]["address"])
                node_list.append(node_item)

        return node_list

    def _check_api_reply(self, api_response, noerror_only=False):
        if noerror_only:
            # Checks if none of the replies has an error
            return lin_drv.all_api_responses_no_error(api_response)
        else:
            # Check if all replies are success
            return lin_drv.all_api_responses_success(api_response)

    def _copy_vol_to_image(self, context, image_service, image_meta, rsc_path,
                           volume):

        return volume_utils.upload_volume(context,
                                          image_service,
                                          image_meta,
                                          rsc_path,
                                          volume)

    #
    # Snapshot
    #
    def create_snapshot(self, snapshot):
        snap_name = self._snapshot_name_from_cinder_snapshot(snapshot)
        rsc_name = self._drbd_resource_name_from_cinder_snapshot(snapshot)

        snap_reply = self._api_snapshot_create(drbd_rsc_name=rsc_name,
                                               snapshot_name=snap_name)

        if not snap_reply:
            msg = 'ERROR creating a LINSTOR snapshot {}'.format(snap_name)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(msg)

    def delete_snapshot(self, snapshot):
        snapshot_name = self._snapshot_name_from_cinder_snapshot(snapshot)
        rsc_name = self._drbd_resource_name_from_cinder_snapshot(snapshot)

        snap_reply = self._api_snapshot_delete(rsc_name, snapshot_name)

        if not snap_reply:
            msg = 'ERROR deleting a LINSTOR snapshot {}'.format(snapshot_name)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(msg)

        # Delete RD if no other RSC are found
        if not self._get_snapshot_nodes(rsc_name):
            self._api_rsc_dfn_delete(rsc_name)

    def create_volume_from_snapshot(self, volume, snapshot):
        src_rsc_name = self._drbd_resource_name_from_cinder_snapshot(snapshot)
        src_snap_name = self._snapshot_name_from_cinder_snapshot(snapshot)
        new_vol_name = self._drbd_resource_name_from_cinder_volume(volume)

        # If no autoplace, manually build a cluster list
        if self.ap_count == 0:
            diskless_nodes = []
            nodes = []
            for node in self._get_storage_pool():

                if DISKLESS in node['driver_name']:
                    diskless_nodes.append(node['node_name'])
                    continue

                # Filter out controller node if it is diskless
                if self.diskless and node['node_name'] == self.host_name:
                    continue
                else:
                    nodes.append(node['node_name'])

        reply = self._api_snapshot_resource_restore(src_rsc_name,
                                                    src_snap_name,
                                                    new_vol_name)
        if not reply:
            msg = _('Error on restoring a LINSTOR volume')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Manually add the controller node as a resource if diskless
        if self.diskless:
            reply = self._api_rsc_create(rsc_name=new_vol_name,
                                         node_name=self.host_name,
                                         diskless=self.diskless)

        # Add any other diskless nodes only if not autoplaced
        if self.ap_count == 0 and diskless_nodes:
            for node in diskless_nodes:
                self._api_rsc_create(rsc_name=new_vol_name,
                                     node_name=node,
                                     diskless=True)

        # Upsize if larger volume than original snapshot
        src_rsc_size = int(snapshot['volume_size'])
        new_vol_size = int(volume['size'])

        if new_vol_size > src_rsc_size:

            upsize_target_name = self._is_clean_volume_name(volume['id'],
                                                            DM_VN_PREFIX)
            reply = self._get_api_volume_extend(
                rsc_target_name=upsize_target_name,
                new_size=new_vol_size)

            if not self._check_api_reply(reply, noerror_only=True):
                # Delete failed volume
                failed_volume = {}
                failed_volume['id'] = volume['id']
                self.delete_volume(failed_volume)

                msg = _('Error on extending LINSTOR resource size')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

    def create_volume(self, volume):

        # Check for Storage Pool List
        sp_data = self._get_storage_pool()
        rsc_size = volume['size']

        # No existing Storage Pools found
        if not sp_data:

            # Check for Nodes
            node_list = self._get_nodes()

            if not node_list:
                msg = _('No LINSTOR resource nodes available / configured')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Create Storage Pool
            spd_list = self._get_spd()

            if spd_list:
                spd_name = spd_list[0]

            for node in node_list:

                node_driver = None
                for sp in sp_data:
                    if sp['node_name'] == node['node_name']:
                        node_driver = sp['driver_name']

                sp_reply = self._api_storage_pool_create(
                    node_name=node['node_name'],
                    storage_pool_name=spd_name,
                    storage_driver=node_driver,
                    driver_pool_name=self.default_vg_name)

                if not self._check_api_reply(sp_reply, noerror_only=True):
                    msg = _('Could not create a LINSTOR storage pool')
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        # Check for RD
        # If Retyping from another volume, use parent/origin uuid
        # as a name source
        if (volume['migration_status'] is not None and
                str(volume['migration_status']).find('success') == -1):
            src_name = str(volume['migration_status']).split(':')[1]
            rsc_name = self._is_clean_volume_name(str(src_name),
                                                  DM_VN_PREFIX)
        else:
            rsc_name = self._is_clean_volume_name(volume['id'],
                                                  DM_VN_PREFIX)

        # Create a New RD
        rsc_dfn_reply = self._api_rsc_dfn_create(rsc_name)

        if not self._check_api_reply(rsc_dfn_reply, noerror_only=True):
            msg = _("Error creating a LINSTOR resource definition")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Create a New VD
        vd_size = self._vol_size_to_linstor(rsc_size)
        vd_reply = self._api_volume_dfn_create(rsc_name=rsc_name,
                                               size=int(vd_size))

        if not self._check_api_reply(vd_reply, noerror_only=True):
            msg = _("Error creating a LINSTOR volume definition")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Create LINSTOR Resources
        ctrl_in_sp = False
        for node in sp_data:
            # Check if controller is in the pool
            if node['node_name'] == self.host_name:
                ctrl_in_sp = True

        # Use autoplace to deploy if set
        if self.ap_count:
            try:
                self._api_rsc_autoplace(rsc_name=rsc_name)

            except Exception:
                msg = _("Error creating autoplaces LINSTOR resource(s)")
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        # Otherwise deploy across the entire cluster
        else:
            for node in sp_data:
                # Deploy resource on each node
                if DISKLESS in node['driver_name']:
                    diskless = True
                else:
                    diskless = False

                rsc_reply = self._api_rsc_create(rsc_name=rsc_name,
                                                 node_name=node['node_name'],
                                                 diskless=diskless)

                if not self._check_api_reply(rsc_reply, noerror_only=True):
                    msg = _("Error creating a LINSTOR resource")
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        # If the controller is diskless and not in the pool, create a diskless
        # resource on it
        if not ctrl_in_sp and self.diskless:
            rsc_reply = self._api_rsc_create(rsc_name=rsc_name,
                                             node_name=self.host_name,
                                             diskless=True)

            if not self._check_api_reply(rsc_reply, noerror_only=True):
                msg = _("Error creating a LINSTOR controller resource")
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        return {}

    def delete_volume(self, volume):
        drbd_rsc_name = self._drbd_resource_name_from_cinder_volume(volume)
        rsc_list_reply = self._get_api_resource_list()
        diskful_nodes = self._get_snapshot_nodes(drbd_rsc_name)
        diskless_nodes = self._get_diskless_nodes(drbd_rsc_name)

        # If autoplace was used, use Resource class
        if self.ap_count:

            rsc_reply = self._api_rsc_auto_delete(drbd_rsc_name)
            if not rsc_reply:
                msg = _("Error deleting an autoplaced LINSTOR resource")
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        # Delete all resources in a cluster manually if not autoplaced
        else:
            if rsc_list_reply:
                # Remove diskless nodes first
                if diskless_nodes:
                    for node in diskless_nodes:
                        rsc_reply = self._api_rsc_delete(
                            node_name=node,
                            rsc_name=drbd_rsc_name)
                        if not self._check_api_reply(rsc_reply,
                                                     noerror_only=True):
                            msg = _("Error deleting a diskless LINSTOR rsc")
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)

                # Remove diskful nodes
                if diskful_nodes:
                    for node in diskful_nodes:
                        rsc_reply = self._api_rsc_delete(
                            node_name=node,
                            rsc_name=drbd_rsc_name)
                        if not self._check_api_reply(rsc_reply,
                                                     noerror_only=True):
                            msg = _("Error deleting a LINSTOR resource")
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)

                # Delete VD
                vd_reply = self._api_volume_dfn_delete(drbd_rsc_name, 0)
                if not vd_reply:
                    if not self._check_api_reply(vd_reply):
                        msg = _("Error deleting a LINSTOR volume definition")
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

                # Delete RD
                # Will fail if snapshot exists but expected
                self._api_rsc_dfn_delete(drbd_rsc_name)

        return True

    def extend_volume(self, volume, new_size):
        rsc_target_name = self._is_clean_volume_name(volume['id'],
                                                     DM_VN_PREFIX)

        extend_reply = self._get_api_volume_extend(rsc_target_name, new_size)

        if not self._check_api_reply(extend_reply, noerror_only=True):
            msg = _("ERROR extending a LINSTOR volume")
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def create_cloned_volume(self, volume, src_vref):
        temp_id = self._clean_uuid()
        snapshot = {}
        snapshot['id'] = temp_id
        snapshot['volume_id'] = src_vref['id']
        snapshot['volume_size'] = src_vref['size']

        self.create_snapshot(snapshot)

        self.create_volume_from_snapshot(volume, snapshot)

        self.delete_snapshot(snapshot)

    def copy_image_to_volume(self, context, volume, image_service, image_id):
        # self.create_volume(volume) already called by Cinder, and works
        full_rsc_name = self._drbd_resource_name_from_cinder_volume(volume)

        # This creates a LINSTOR volume from the source image
        image_utils.fetch_to_raw(context,
                                 image_service,
                                 image_id,
                                 str(self._get_rsc_path(full_rsc_name)),
                                 self.default_blocksize,
                                 size=volume['size'])
        return {}

    def copy_volume_to_image(self, context, volume, image_service, image_meta):
        full_rsc_name = self._drbd_resource_name_from_cinder_volume(volume)
        rsc_path = str(self._get_rsc_path(full_rsc_name))
        self._copy_vol_to_image(context,
                                image_service,
                                image_meta,
                                rsc_path,
                                volume)
        return {}

    # Not supported currently
    def migrate_volume(self, ctxt, volume, host, thin=False, mirror_count=0):
        return (False, None)

    def check_for_setup_error(self):
        msg = None
        if linstor is None:
            msg = _('Linstor python package not found')

        if msg is not None:
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

    def create_export(self, context, volume, connector):
        pass

    def ensure_export(self, context, volume):
        pass

    def initialize_connection(self, volume, connector, **kwargs):
        pass

    def remove_export(self, context, volume):
        pass

    def terminate_connection(self, volume, connector, **kwargs):
        pass


# Class with iSCSI interface methods
@interface.volumedriver
class LinstorIscsiDriver(LinstorBaseDriver):
    """Cinder iSCSI driver that uses LINSTOR for storage."""

    def __init__(self, *args, **kwargs):
        super(LinstorIscsiDriver, self).__init__(*args, **kwargs)

        # iSCSI target_helper
        if 'h_name' in kwargs:
            self.helper_name = kwargs.get('h_name')
            self.helper_driver = self.helper_name
            self.target_driver = None
        else:
            self.helper_name = self.configuration.safe_get('iscsi_helper')
            self.helper_driver = self.target_mapping[self.helper_name]
            self.target_driver = importutils.import_object(
                self.helper_driver,
                configuration=self.configuration,
                db=self.db,
                executor=self._execute)

        LOG.info('START: LINSTOR DRBD driver %s', self.helper_name)

    def get_volume_stats(self, refresh=False):
        data = self._get_volume_stats()
        data["storage_protocol"] = 'iSCSI'
        data["pools"][0]["location_info"] = (
            'LinstorIscsiDriver:' + data["pools"][0]["location_info"])

        return data

    def ensure_export(self, context, volume):
        volume_path = self._get_local_path(volume)

        return self.target_driver.ensure_export(
            context,
            volume,
            volume_path)

    def create_export(self, context, volume, connector):
        volume_path = self._get_local_path(volume)

        export_info = self.target_driver.create_export(
            context,
            volume,
            volume_path)

        return {'provider_location': export_info['location'],
                'provider_auth': export_info['auth'], }

    def remove_export(self, context, volume):

        return self.target_driver.remove_export(context, volume)

    def initialize_connection(self, volume, connector, **kwargs):

        return self.target_driver.initialize_connection(volume, connector)

    def validate_connector(self, connector):

        return self.target_driver.validate_connector(connector)

    def terminate_connection(self, volume, connector, **kwargs):

        return self.target_driver.terminate_connection(volume,
                                                       connector,
                                                       **kwargs)


# Class with DRBD transport mode
@interface.volumedriver
class LinstorDrbdDriver(LinstorBaseDriver):
    """Cinder DRBD driver that uses LINSTOR for storage."""

    def __init__(self, *args, **kwargs):
        super(LinstorDrbdDriver, self).__init__(*args, **kwargs)

    def _return_drbd_config(self, volume):
        full_rsc_name = self._drbd_resource_name_from_cinder_volume(volume)
        rsc_path = self._get_rsc_path(full_rsc_name)
        return {
            'driver_volume_type': 'local',
            'data': {
                "device_path": str(rsc_path)
            }
        }

    def _node_in_sp(self, node_name):
        for pool in self._get_storage_pool():
            if pool['node_name'] == node_name:
                return True
        return False

    def get_volume_stats(self, refresh=False):
        data = self._get_volume_stats()
        data["storage_protocol"] = 'DRBD'
        data["pools"][0]["location_info"] = 'LinstorDrbdDriver:{}'.format(
            data["pools"][0]["location_info"])

        return data

    def initialize_connection(self, volume, connector, **kwargs):
        node_name = connector['host']
        if not self._node_in_sp(connector['host']):

            full_rsc_name = self._drbd_resource_name_from_cinder_volume(volume)
            rsc_reply = self._api_rsc_create(rsc_name=full_rsc_name,
                                             node_name=node_name,
                                             diskless=True)
            if not self._check_api_reply(rsc_reply, noerror_only=True):
                msg = _('Error on creating LINSTOR Resource')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

        return self._return_drbd_config(volume)

    def terminate_connection(self, volume, connector, **kwargs):
        if connector:
            node_name = connector['host']
            if not self._node_in_sp(connector['host']):
                rsc_name = self._drbd_resource_name_from_cinder_volume(volume)
                rsc_reply = self._api_rsc_delete(rsc_name=rsc_name,
                                                 node_name=node_name)
                if not self._check_api_reply(rsc_reply, noerror_only=True):
                    msg = _('Error on deleting LINSTOR Resource')
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

    def create_export(self, context, volume, connector):

        return self._return_drbd_config(volume)

    def ensure_export(self, context, volume):

        return self._return_drbd_config(volume)

    def remove_export(self, context, volume):
        pass




# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for the API endpoint."""

import six
from six.moves import http_client
import webob


class FakeHttplibSocket(object):
    """A fake socket implementation for http_client.HTTPResponse, trivial."""
    def __init__(self, response_string):
        self.response_string = response_string
        self._buffer = six.StringIO(response_string)

    def makefile(self, _mode, _other):
        """Returns the socket's internal buffer."""
        return self._buffer


class FakeHttplibConnection(object):
    """A fake http_client.HTTPConnection for boto.

    requests made via this connection actually get translated and routed into
    our WSGI app, we then wait for the response and turn it back into
    the http_client.HTTPResponse that boto expects.
    """
    def __init__(self, app, host, is_secure=False):
        self.app = app
        self.host = host

    def request(self, method, path, data, headers):
        req = webob.Request.blank(path)
        req.method = method
        req.body = data
        req.headers = headers
        req.headers['Accept'] = 'text/html'
        req.host = self.host
        # Call the WSGI app, get the HTTP response
        resp = str(req.get_response(self.app))
        # For some reason, the response doesn't have "HTTP/1.0 " prepended; I
        # guess that's a function the web server usually provides.
        resp = "HTTP/1.0 %s" % resp
        self.sock = FakeHttplibSocket(resp)
        self.http_response = http_client.HTTPResponse(self.sock)
        # NOTE(vish): boto is accessing private variables for some reason
        self._HTTPConnection__response = self.http_response
        self.http_response.begin()

    def getresponse(self):
        return self.http_response

    def getresponsebody(self):
        return self.sock.response_string

    def close(self):
        """Required for compatibility with boto/tornado."""
        pass




# Copyright 2012, Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit Tests for cinder.scheduler.rpcapi."""

from datetime import datetime
from unittest import mock

import ddt

from cinder import exception
from cinder import objects
from cinder.scheduler import rpcapi as scheduler_rpcapi
from cinder.tests.unit.backup import fake_backup
from cinder.tests.unit import fake_constants
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test


@ddt.ddt
class SchedulerRPCAPITestCase(test.RPCAPITestCase):
    def setUp(self):
        super(SchedulerRPCAPITestCase, self).setUp()
        self.rpcapi = scheduler_rpcapi.SchedulerAPI
        self.base_version = '3.0'
        self.volume_id = fake_constants.VOLUME_ID
        self.fake_volume = fake_volume.fake_volume_obj(
            self.context, expected_attrs=['metadata', 'admin_metadata',
                                          'glance_metadata'])
        self.fake_snapshot = fake_snapshot.fake_snapshot_obj(
            self.context)
        self.fake_rs_obj = objects.RequestSpec.from_primitives({})
        self.fake_rs_dict = {'volume_id': self.volume_id}
        self.fake_fp_dict = {'availability_zone': 'fake_az'}
        self.fake_backup_dict = fake_backup.fake_backup_obj(self.context)

    @ddt.data('3.0', '3.3')
    @mock.patch('oslo_messaging.RPCClient.can_send_version')
    def test_update_service_capabilities(self, version, can_send_version):
        can_send_version.side_effect = lambda x: x == version
        self._test_rpc_api('update_service_capabilities',
                           rpc_method='cast',
                           service_name='fake_name',
                           host='fake_host',
                           cluster_name='cluster_name',
                           capabilities={},
                           fanout=True,
                           version=version,
                           timestamp='123')
        can_send_version.assert_called_once_with('3.3')

    @ddt.data('3.0', '3.10')
    @mock.patch('oslo_messaging.RPCClient.can_send_version')
    def test_create_volume(self, version, can_send_version):
        can_send_version.side_effect = lambda x: x == version
        create_worker_mock = self.mock_object(self.fake_volume,
                                              'create_worker')
        self._test_rpc_api('create_volume',
                           rpc_method='cast',
                           volume=self.fake_volume,
                           snapshot_id=fake_constants.SNAPSHOT_ID,
                           image_id=fake_constants.IMAGE_ID,
                           backup_id=fake_constants.BACKUP_ID,
                           request_spec=self.fake_rs_obj,
                           filter_properties=self.fake_fp_dict)
        create_worker_mock.assert_called_once()
        can_send_version.assert_called_once_with('3.10')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_create_snapshot(self, can_send_version_mock):
        self._test_rpc_api('create_snapshot',
                           rpc_method='cast',
                           volume='fake_volume',
                           snapshot='fake_snapshot',
                           backend='fake_backend',
                           request_spec={'snapshot_id': self.fake_snapshot.id},
                           filter_properties=None)

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_create_snapshot_capped(self, can_send_version_mock):
        self.assertRaises(exception.ServiceTooOld,
                          self._test_rpc_api,
                          'create_snapshot',
                          rpc_method='cast',
                          volume=self.fake_volume,
                          snapshot=self.fake_snapshot,
                          backend='fake_backend',
                          request_spec=self.fake_rs_obj,
                          version='3.5')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=True)
    def test_manage_existing_snapshot(self, can_send_version_mock):
        self._test_rpc_api('manage_existing_snapshot',
                           rpc_method='cast',
                           volume='fake_volume',
                           snapshot='fake_snapshot',
                           ref='fake_ref',
                           request_spec={'snapshot_id': self.fake_snapshot.id},
                           filter_properties=None)

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_manage_existing_snapshot_capped(self, can_send_version_mock):
        self.assertRaises(exception.ServiceTooOld,
                          self._test_rpc_api,
                          'manage_existing_snapshot',
                          rpc_method='cast',
                          volume=self.fake_volume,
                          snapshot=self.fake_snapshot,
                          ref='fake_ref',
                          request_spec={'snapshot_id': self.fake_snapshot.id,
                                        'ref': 'fake_ref'},
                          filter_properties=None,
                          version='3.10')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_notify_service_capabilities_backend(self, can_send_version_mock):
        """Test sending new backend by RPC instead of old host parameter."""
        capabilities = {'host': 'fake_host',
                        'total': '10.01', }
        with mock.patch('oslo_utils.timeutils.utcnow',
                        return_value=datetime(1970, 1, 1)):
            self._test_rpc_api('notify_service_capabilities',
                               rpc_method='cast',
                               service_name='fake_name',
                               backend='fake_host',
                               capabilities=capabilities,
                               timestamp='1970-01-01T00:00:00.000000',
                               version='3.5')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                side_effect=(True, False))
    def test_notify_service_capabilities_host(self, can_send_version_mock):
        """Test sending old host RPC parameter instead of backend."""
        capabilities = {'host': 'fake_host',
                        'total': '10.01', }
        self._test_rpc_api('notify_service_capabilities',
                           rpc_method='cast',
                           service_name='fake_name',
                           server='fake_host',
                           expected_kwargs_diff={'host': 'fake_host'},
                           backend='fake_host',
                           capabilities=capabilities,
                           version='3.1')

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_notify_service_capabilities_capped(self, can_send_version_mock):
        capabilities = {'host': 'fake_host',
                        'total': '10.01', }
        self.assertRaises(exception.ServiceTooOld,
                          self._test_rpc_api,
                          'notify_service_capabilities',
                          rpc_method='cast',
                          service_name='fake_name',
                          backend='fake_host',
                          server='fake_host',
                          # ignore_for_method=['host'],
                          # ignore_for_rpc=['backend'],
                          capabilities=capabilities,
                          version='3.1')

    @mock.patch('oslo_messaging.RPCClient.can_send_version')
    def test_migrate_volume(self, can_send_version):
        create_worker_mock = self.mock_object(self.fake_volume,
                                              'create_worker')
        self._test_rpc_api('migrate_volume',
                           rpc_method='cast',
                           backend='host',
                           force_copy=True,
                           request_spec='fake_request_spec',
                           filter_properties='filter_properties',
                           volume=self.fake_volume,
                           version='3.3')
        create_worker_mock.assert_not_called()

    def test_retype(self):
        self._test_rpc_api('retype',
                           rpc_method='cast',
                           request_spec=self.fake_rs_dict,
                           filter_properties=self.fake_fp_dict,
                           volume=self.fake_volume)

    def test_manage_existing(self):
        self._test_rpc_api('manage_existing',
                           rpc_method='cast',
                           request_spec=self.fake_rs_dict,
                           filter_properties=self.fake_fp_dict,
                           volume=self.fake_volume)

    @mock.patch('oslo_messaging.RPCClient.can_send_version',
                return_value=False)
    def test_extend_volume_capped(self, can_send_version_mock):
        self.assertRaises(exception.ServiceTooOld,
                          self._test_rpc_api,
                          'extend_volume',
                          rpc_method='cast',
                          request_spec='fake_request_spec',
                          filter_properties='filter_properties',
                          volume=self.fake_volume,
                          new_size=4,
                          reservations=['RESERVATIONS'],
                          version='3.0')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', return_value=True)
    def test_extend_volume(self, can_send_version_mock):
        create_worker_mock = self.mock_object(self.fake_volume,
                                              'create_worker')
        self._test_rpc_api('extend_volume',
                           rpc_method='cast',
                           request_spec='fake_request_spec',
                           filter_properties='filter_properties',
                           volume=self.fake_volume,
                           new_size=4,
                           reservations=['RESERVATIONS'])
        create_worker_mock.assert_not_called()

    def test_get_pools(self):
        self._test_rpc_api('get_pools',
                           rpc_method='call',
                           filters=None,
                           retval=[{
                               'name': 'fake_pool',
                               'capabilities': {},
                           }])

    def test_create_group(self):
        self._test_rpc_api('create_group',
                           rpc_method='cast',
                           group='group',
                           group_spec=self.fake_rs_dict,
                           request_spec_list=[self.fake_rs_dict],
                           group_filter_properties=[self.fake_fp_dict],
                           filter_properties_list=[self.fake_fp_dict])

    @ddt.data(('work_cleanup', 'myhost', None),
              ('work_cleanup', 'myhost', 'mycluster'),
              ('do_cleanup', 'myhost', None),
              ('do_cleanup', 'myhost', 'mycluster'))
    @ddt.unpack
    @mock.patch('cinder.rpc.get_client')
    def test_cleanup(self, method, host, cluster, get_client):
        cleanup_request = objects.CleanupRequest(self.context,
                                                 host=host,
                                                 cluster_name=cluster)
        rpcapi = scheduler_rpcapi.SchedulerAPI()
        getattr(rpcapi, method)(self.context, cleanup_request)

        prepare = get_client.return_value.prepare

        prepare.assert_called_once_with(
            version='3.4')
        rpc_call = 'cast' if method == 'do_cleanup' else 'call'
        getattr(prepare.return_value, rpc_call).assert_called_once_with(
            self.context, method, cleanup_request=cleanup_request)

    @ddt.data('do_cleanup', 'work_cleanup')
    def test_cleanup_too_old(self, method):
        cleanup_request = objects.CleanupRequest(self.context)
        rpcapi = scheduler_rpcapi.SchedulerAPI()
        with mock.patch.object(rpcapi.client, 'can_send_version',
                               return_value=False) as can_send_mock:
            self.assertRaises(exception.ServiceTooOld,
                              getattr(rpcapi, method),
                              self.context,
                              cleanup_request)
            can_send_mock.assert_called_once_with('3.4')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', mock.Mock())
    def test_set_log_levels(self):
        service = objects.Service(self.context, host='host1')
        self._test_rpc_api('set_log_levels',
                           rpc_method='cast',
                           server=service.host,
                           service=service,
                           log_request='log_request',
                           version='3.7')

    @mock.patch('oslo_messaging.RPCClient.can_send_version', mock.Mock())
    def test_get_log_levels(self):
        service = objects.Service(self.context, host='host1')
        self._test_rpc_api('get_log_levels',
                           rpc_method='call',
                           server=service.host,
                           service=service,
                           log_request='log_request',
                           version='3.7')

    @mock.patch('oslo_messaging.RPCClient.can_send_version')
    def test_create_backup(self, can_send_version):
        self._test_rpc_api('create_backup',
                           rpc_method='cast',
                           backup=self.fake_backup_dict)

        can_send_version.assert_called_once_with('3.12')
